# tml - 0.91 (beta)
# (c) Piotr Kowalewski, August-September 2024
from hpprime import eval as ppleval, strblit2, fillrect, keyboard
from cas import get_key

class tml:
    def __init__(self, dark_mode=False, tab_size=4):
        # Initialize terminal dimensions and properties
        self.columns = 64
        self.rows = 24
        self.cursor_x = 0
        self.cursor_y = 0
        self.char_width = 5
        self.char_height = 10
        self.width = self.columns * self.char_width
        self.height = self.rows * self.char_height
        self.back_color = 0
        self.tab_size = tab_size
        self.char_mapping = {i: i-32 for i in range(32, 127)}
        self.is_alpha = False
        self.is_shift = False
        self.last_keyboard_state = keyboard()
        ppleval('G1:=AFiles("font_5x10.png")')
        if dark_mode:
            ppleval('INVERT_P(G1)')
        else:
            self.back_color = rgb(255,255,255)
        self.clear()

    def output(self, *args, sep=' ', end='\n'):
        # Output text based on the provided arguments
        text = sep.join(str(arg) for arg in args) + end
        for char in text:
            self._put_char(char)

    def clear(self):
        # Clear the entire screen and reset cursor position
        fillrect(0, 0, 0, self.width, self.height, self.back_color, self.back_color)
        self.cursor_x = 0
        self.cursor_y = 0

    def set_cursor(self, x, y):
        # Set the cursor position (if within bounds)
        if 0 <= x < self.columns and 0 <= y < self.rows:
            self.cursor_x = x
            self.cursor_y = y
        else:
            raise ValueError("Cursor position out of bounds")

    def input(self, prompt=None, alpha=True, shift=False):
        # Custom input method to gather text input from the user
        self.is_alpha = alpha
        self.is_shift = shift
        if prompt:
            self.output(prompt, end='')
        pos = self.cursor_x
        input_string = ''
        self._invert_cursor()
        try:
            while True:
                char = self.read_key()
                if char == None:
                    continue
                if char == '\n':
                    self._invert_cursor()
                    self._put_char('\n')
                    break
                elif char == '\b':
                    if self.cursor_x > pos:
                        self._invert_cursor()
                        input_string = input_string[:-1]
                        self.cursor_x -= 1
                        self._draw_char(' ')
                        self._invert_cursor()
                elif char == '\Clr':
                    self._invert_cursor()
                    while self.cursor_x > pos:
                        self.cursor_x -= 1
                        self._draw_char(' ')
                    self._invert_cursor()
                    input_string = ''
                elif char == '\Esc':
                    raise KeyboardInterrupt
                else:
                    if self.cursor_x+1 < self.columns:
                        self._invert_cursor()
                        input_string += char
                        self._put_char(char)
                        self._invert_cursor()
        except KeyboardInterrupt:
            self._invert_cursor()
            while self.cursor_x > pos:
                self.cursor_x -= 1
                self._draw_char(' ')
            self._put_char('\n')
            input_string = ''
        return input_string

    def read_key(self):
        # Read key input from the keyboard
        while True:
            current_state = keyboard()
            changed_keys = current_state ^ self.last_keyboard_state  # XOR to find changes
            if changed_keys:  # Check if there are any changes
                self.last_keyboard_state = current_state  # Update the last known state
                for key_index in range(52):  # 52 possible keys
                    if changed_keys & (1 << key_index):  # Check if the i-th key state has changed
                        if current_state & (1 << key_index):  # Check if the i-th key is pressed
                            get_key()
                            if key_index == 36:
                                self.is_alpha = not self.is_alpha
                            elif key_index == 41:
                                self.is_shift = not self.is_shift
                            else:
                                # Map key index to respective character
                                key_map = {
                                    # no modifiers, Alpha, Shift, Alpha+Shift
                                    4: ['\Esc',None,'\Clr','\Clr'],
                                    14: [None,'a',None,'A'],
                                    15: [None,'b',None,'B'],
                                    16: [None,'c',None,'C'],
                                    17: [None,'d',None,'D'],
                                    18: [None,'e',None,'E'],
                                    19: ['\b','\b','\b','\b'],
                                    20: ['^','f',None,'F'],
                                    21: [None,'g',None,'G'],
                                    22: [None,'h',None,'H'],
                                    23: [None,'i',None,'I'],
                                    24: [None,'j',None,'J'],
                                    25: [None,'k',None,'K'],
                                    26: [None,'l',None,'L'],
                                    27: [None,'m','|','M'],
                                    28: [None,'n',"'",'N'],
                                    29: [',','o',None,'O'],
                                    30: ['\n','\n','\n','\n'],
                                    31: [None,'p',None,'P'],
                                    32: ['7','q','&','Q'],
                                    33: ['8','r',None,'R'],
                                    34: ['9','s','!','S'],
                                    35: ['/','t',None,'T'],
                                    37: ['4','u','$','U'],
                                    38: ['5','v','%','V'],
                                    39: ['6','w','^','W'],
                                    40: ['*','x','(','X'],
                                    42: ['1','y',None,'Y'],
                                    43: ['2','z','@','Z'],
                                    44: ['3','#','?','#'],
                                    45: ['-',':',')',':'],
                                    47: ['0','"',None,'"'],
                                    48: ['.','.','=','.'],
                                    49: [' ',' ','_','_'],
                                    50: ['+',';','\\',None]
                                    }
                                modifiers_state = (self.is_shift << 1) | self.is_alpha
                                return key_map.get(key_index, [None,None,None,None])[modifiers_state]
            ppleval('WAIT(0.001)')

    def _put_char(self, char):
        # Place character on the screen and manage cursor position
        if char == '\n':
            self.cursor_x = 0
            self.cursor_y += 1
            self._end_of_screen_check()
        elif char == '\t':
            # Handle tabulation based on tab size
            spaces_to_next_tab = self.tab_size - (self.cursor_x % self.tab_size)
            self.cursor_x += spaces_to_next_tab
            if self.cursor_x >= self.columns:
                self.cursor_x = 0
                self.cursor_y += 1
                self._end_of_screen_check()
        else:
            self._draw_char(char)
            self.cursor_x += 1
            if self.cursor_x >= self.columns:
                self.cursor_x = 0
                self.cursor_y += 1
                self._end_of_screen_check()

    def _draw_char(self, char):
        # Draw character to the screen using a bitmap
        char_code = ord(char)
        if char_code in self.char_mapping:
            index = self.char_mapping[char_code]
            char_x = index * self.char_width
            strblit2(0, self.cursor_x * self.char_width, self.cursor_y * self.char_height, self.char_width, self.char_height, 1, char_x, 0, self.char_width, self.char_height)

    def _end_of_screen_check(self):
        # Check if the cursor has moved past the bottom edge of the screen
        if self.cursor_y >= self.rows:
            self._scroll_up()
    
    def _scroll_up(self):
        # Scroll the screen up by one line
        strblit2(0, 0, 0, self.width, self.height - self.char_height, 0, 0, self.char_height, self.width, self.height - self.char_height)
        fillrect(0, 0, self.height - self.char_height, self.width, self.char_height, self.back_color, self.back_color)
        self.cursor_y -= 1

    def _invert_cursor(self):
        # Invert the cursor display
        ppleval('INVERT_P(%d,%d,%d,%d)' % (self.cursor_x * self.char_width, self.cursor_y * self.char_height, self.cursor_x * self.char_width + self.char_width - 1, self.cursor_y * self.char_height + self.char_height - 1))

# public methods
def rgb(r,g,b):
    return (r*256+g)*256+b